"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.mockEndpoint = void 0;
const tslib_1 = require("tslib");
const api_client_js_1 = require("./api-client.js");
let mocks;
/** Mock a coalesce API endpoint, providing your own implementation for the API request.
 
 * Example:
 * ```
 * mockEndpoint(
      "/Course/list",
      (req) => ({
        wasSuccessful: true,
        list: [
          { courseId: 1, title: "CS 101" },
          { courseId: 2, title: "CS 102" },
        ],
      })
    )
 * ```
 *
 * Example:
 * ```
 * const fn = mockEndpoint(
      new Student().$metadata.methods.getName,
      vitest.fn((req) => ({
        wasSuccessful: true,
        object: "Steve",
      }))
    )
    // Do testing ...
    expect(fn).toHaveBeenCalledOnce();
    fn.destroy();
 * ```
 */
function mockEndpoint(endpoint, mock) {
    if (!mocks) {
        mocks = [];
        api_client_js_1.AxiosClient.interceptors.request.use((config) => {
            var _a;
            for (let i = mocks.length - 1; i >= 0; i--) {
                const mock = mocks[i];
                if ((typeof mock.endpoint == "string" &&
                    ((_a = config.url) === null || _a === void 0 ? void 0 : _a.toLowerCase().startsWith(mock.endpoint.toLowerCase()))) ||
                    (typeof mock.endpoint == "object" &&
                        "__coalesce_method_meta" in config &&
                        config.__coalesce_method_meta() == mock.endpoint)) {
                    // This request matches a registered mock.
                    // Substitute out the adapter for our mock.
                    config.adapter = function (config) {
                        var _a, _b, _c, _d, _e, _f;
                        return tslib_1.__awaiter(this, void 0, void 0, function* () {
                            const result = yield mock.adapter(config);
                            // If the result is a ListResult and the mock didn't set pagination fields,
                            // fill them in so that server behavior is emulated more accurately without requiring
                            // the user to have to do this tediousness on every mock.
                            if ("list" in result) {
                                (_a = result.page) !== null && _a !== void 0 ? _a : (result.page = 1);
                                (_b = result.totalCount) !== null && _b !== void 0 ? _b : (result.totalCount = (_d = (_c = result.list) === null || _c === void 0 ? void 0 : _c.length) !== null && _d !== void 0 ? _d : 0);
                                (_e = result.pageSize) !== null && _e !== void 0 ? _e : (result.pageSize = Math.max(10, result.totalCount));
                                (_f = result.pageCount) !== null && _f !== void 0 ? _f : (result.pageCount = Math.ceil(result.totalCount / result.pageSize));
                            }
                            const response = {
                                status: result.wasSuccessful ? 200 : 400,
                                statusText: result.wasSuccessful ? "OK" : "Bad Request",
                                data: result,
                                config,
                                headers: { "Content-Type": "application/json" },
                            };
                            const validateStatus = response.config.validateStatus;
                            if (!validateStatus || validateStatus(response.status)) {
                                return response;
                            }
                            else {
                                var error = new Error("Request failed with status code " + response.status);
                                error.isAxiosError = true;
                                error.response = response;
                                throw error;
                            }
                        });
                    };
                    return config;
                }
            }
            return config;
        });
    }
    const mockData = {
        adapter: mock,
        endpoint,
    };
    mocks.push(mockData);
    // Return the passed in mock so it can be wrapped in a `vitest.fn`
    // so the user can spy on it.
    return Object.assign(mock, {
        /** Tear down the mock, removing the handler from the global axios adapter. */
        destroy() {
            mocks.splice(mocks.indexOf(mockData), 1);
        },
    });
}
exports.mockEndpoint = mockEndpoint;
