import type { ModelType, PropertyOrName, PropNames, ModelCollectionNavigationProperty, ModelCollectionValue, Service, Property } from "./metadata.js";
import { ModelApiClient, DataSourceParameters, ServiceApiClient } from "./api-client.js";
import { type Model } from "./model.js";
import { type Indexable, type DeepPartial, type VueInstance } from "./util.js";
import type { DebounceSettings } from "lodash";
export type { DeepPartial } from "./util.js";
import type * as apiClient from "./api-client.js";
export declare abstract class ViewModel<TModel extends Model<ModelType> = any, TApi extends ModelApiClient<TModel> = any, TPrimaryKey extends string | number = any> implements Model<TModel["$metadata"]> {
    /** The metadata representing the type of data that this ViewModel handles. */
    readonly $metadata: TModel["$metadata"];
    /** Instance of an API client for the model through which direct, stateless API requests may be made. */
    readonly $apiClient: TApi;
    /** Static lookup of all generated ViewModel types. */
    static typeLookup: ViewModelTypeLookup | null;
    /**
     * A number unique among all ViewModel instances that will be unchanged for the instance's lifetime.
     * Can be used to uniquely identify objects with `:key` in Vue components.
     *
     * Especially useful with `transition-group` where some elements of iteration may be unsaved
     * and therefore not yet have a `$primaryKey`.
     */
    readonly $stableId: number;
    /**
     * Gets or sets the primary key of the ViewModel's data.
     */
    get $primaryKey(): TPrimaryKey;
    set $primaryKey(val: TPrimaryKey);
    /**
     * Returns true if the values of the savable data properties of this ViewModel
     * have changed since the last load, save, or the last time $isDirty was set to false.
     */
    get $isDirty(): boolean;
    set $isDirty(val: boolean);
    $setPropDirty(propName: PropNames<TModel["$metadata"]>, dirty?: boolean, triggerAutosave?: boolean): void;
    $getPropDirty(propName: PropNames<TModel["$metadata"]>): boolean;
    /** The parameters that will be passed to `/get`, `/save`, and `/delete` calls. */
    get $params(): DataSourceParameters;
    set $params(val: DataSourceParameters);
    /** Wrapper for `$params.dataSource` */
    get $dataSource(): import("./model.js").DataSource<import("./metadata.js").DataSourceType> | null | undefined;
    set $dataSource(val: import("./model.js").DataSource<import("./metadata.js").DataSourceType> | null | undefined);
    /** Wrapper for `$params.includes` */
    get $includes(): string | null | undefined;
    set $includes(val: string | null | undefined);
    $addRule(prop: string | Property, identifier: string, rule: (val: any) => true | string): void;
    $removeRule(prop: string | Property, identifier: string): void;
    /** Returns an array of all applicable rules for the given property
      in accordance with:
      - The rules defined for the model's properties' metadata,
      - Custom rules added by calling `$addRule`
      - Any rules that where ignored by calling `this.$removeRule`,*/
    $getRules(prop: string | Property): ((val: any) => string | true)[] | undefined;
    /** Returns a generator that provides all error messages for the current model
      in accordance with:
      - The rules defined for the model's properties' metadata,
      - Custom rules added by calling `$addRule`
      - Any rules that where ignored by calling `this.$removeRule`,
    */
    $getErrors(prop?: string | Property): Generator<string, void, unknown>;
    /** True if `this.$getErrors()` is returning at least one error. */
    get $hasError(): boolean;
    /**
     * A function for invoking the `/get` endpoint, and a set of properties about the state of the last call.
     */
    get $load(): apiClient.ItemApiState<[id?: TPrimaryKey | undefined], TModel>;
    /** Whether or not to reload the ViewModel's `$data` with the response received from the server after a call to .$save(). */
    $loadResponseFromSaves: boolean;
    /** Whether to save the whole object, or only properties which are dirty.
     * - `surgical` - save only properties that are flagged as dirty (default).
     * - `whole` - save all properties on the object.
     */
    $saveMode: "surgical" | "whole";
    /** When `$save.isLoading == true`, contains the properties of the model currently being saved by `$save` (including autosaves).
     *
     * Does not include non-dirty properties even if `$saveMode == 'whole'`.
     */
    get $savingProps(): ReadonlySet<string>;
    set $savingProps(value: ReadonlySet<string>);
    /**
     * A function for invoking the `/save` endpoint, and a set of properties about the state of the last call.
     */
    get $save(): apiClient.ItemApiState<[overrideProps?: Partial<TModel> | undefined], TModel>;
    /**
     * A function for invoking the `/bulkSave` endpoint, and a set of properties about the state of the last call.
     *
     * A bulk save will save/delete this entity and all reachable relationships in a single round trip and database transaction.
     */
    get $bulkSave(): apiClient.ItemApiState<[options?: BulkSaveOptions | undefined], TModel>;
    /**
     * Loads data from the provided model into the current ViewModel,
     * and then clears the $isDirty flag if `isCleanData` is not given as `false`.
     *
     * Data is loaded in a surgical fashion that will preserve existing instances
     * of objects and arrays found on navigation properties.
     */
    $loadFromModel(source: DeepPartial<TModel>, isCleanData?: boolean, purgeUnsaved?: boolean): void;
    /**
     * Loads data from the provided model into the current ViewModel, and then clears $isDirty flags.
     *
     * Data is loaded in a surgical fashion that will preserve existing instances
     * of objects and arrays found on navigation properties.
     */
    $loadCleanData(source: DeepPartial<TModel>, purgeUnsaved?: boolean): void;
    /**
     * Loads data from the provided model into the current ViewModel.
     * Does not clear $isDirty flags.
     *
     * Data is loaded in a surgical fashion that will preserve existing instances
     * of objects and arrays found on navigation properties.
     */
    $loadDirtyData(source: DeepPartial<TModel>): void;
    /**
     * A function for invoking the `/delete` endpoint, and a set of properties about the state of the last call.
     */
    get $delete(): apiClient.ItemApiState<[], TModel>;
    private get _existsOnServer();
    get $isRemoved(): boolean;
    /** Mark this model for deletion in the next `$bulkSave` operation performed on an ancestor.
     *
     * If this item has a parent collection, it will be immediately removed from that collection.
     * */
    $remove(): void;
    /**
     * Starts auto-saving of the instance when changes to its savable data properties occur.
     * Only usable from Vue setup() or <script setup>. Otherwise, use $startAutoSave().
     * @param options Options to control how the auto-saving is performed.
     */
    $useAutoSave(options?: AutoSaveOptions<this>): void;
    /**
     * Starts auto-saving of the instance when changes to its savable data properties occur.
     * @param vue A Vue instance through which the lifecycle of the watcher will be managed.
     * @param options Options to control how the auto-saving is performed.
     */
    $startAutoSave(vue: VueInstance, options?: AutoSaveOptions<this>): void;
    /** Stops auto-saving if it is currently enabled. */
    $stopAutoSave(): void;
    /** Returns true if auto-saving is currently enabled. */
    get $isAutoSaveEnabled(): boolean | undefined;
    /**
     * Returns a string representation of the object, or one of its properties, suitable for display.
     * @param prop If provided, specifies a property whose value will be displayed.
     * If omitted, the whole object will be represented.
     */
    $display(prop?: PropertyOrName<TModel["$metadata"]>): string | null;
    /**
     * Creates a new instance of an item for the specified child model collection,
     * adds it to that collection, and returns the item.
     * @param prop The name of the collection property, or the metadata representing it.
     */
    $addChild(prop: ModelCollectionNavigationProperty | PropNames<TModel["$metadata"], ModelCollectionNavigationProperty>, initialDirtyData?: any): ViewModel<any, any, any>;
    constructor(
    /** The metadata representing the type of data that this ViewModel handles. */
    $metadata: TModel["$metadata"], 
    /** Instance of an API client for the model through which direct, stateless API requests may be made. */
    $apiClient: TApi, initialDirtyData?: DeepPartial<TModel> | null);
}
export declare abstract class ListViewModel<TModel extends Model<ModelType> = any, TApi extends ModelApiClient<TModel> = ModelApiClient<TModel>, TItem extends ViewModel<TModel, TApi> = ViewModel<TModel, TApi>> {
    /** The metadata representing the type of data that this ViewModel handles. */
    readonly $metadata: TModel["$metadata"];
    /** Instance of an API client for the model through which direct, stateless API requests may be made. */
    readonly $apiClient: TApi;
    /** Static lookup of all generated ListViewModel types. */
    static typeLookup: ListViewModelTypeLookup | null;
    /** The parameters that will be passed to `/list` and `/count` calls. */
    get $params(): {
        page?: number | null | undefined;
        pageSize?: number | null | undefined;
        noCount?: boolean | null | undefined;
        orderBy?: string | null | undefined;
        orderByDescending?: string | null | undefined;
        fields?: string[] | null | undefined;
        search?: string | null | undefined;
        filter?: {
            [fieldName: string]: string | number | boolean | null | undefined;
        } | null | undefined;
        includes?: string | null | undefined;
        dataSource?: {
            readonly $metadata: {
                readonly type: "dataSource";
                readonly isDefault?: true | undefined;
                readonly props: {
                    [x: string]: import("./metadata.js").PrimitiveProperty | import("./metadata.js").DateProperty | import("./metadata.js").EnumProperty;
                };
                readonly name: string;
                readonly displayName: string;
            };
        } | null | undefined;
    };
    set $params(val: {
        page?: number | null | undefined;
        pageSize?: number | null | undefined;
        noCount?: boolean | null | undefined;
        orderBy?: string | null | undefined;
        orderByDescending?: string | null | undefined;
        fields?: string[] | null | undefined;
        search?: string | null | undefined;
        filter?: {
            [fieldName: string]: string | number | boolean | null | undefined;
        } | null | undefined;
        includes?: string | null | undefined;
        dataSource?: {
            readonly $metadata: {
                readonly type: "dataSource";
                readonly isDefault?: true | undefined;
                readonly props: {
                    [x: string]: import("./metadata.js").PrimitiveProperty | import("./metadata.js").DateProperty | import("./metadata.js").EnumProperty;
                };
                readonly name: string;
                readonly displayName: string;
            };
        } | null | undefined;
    });
    /** Wrapper for `$params.dataSource` */
    get $dataSource(): {
        readonly $metadata: {
            readonly type: "dataSource";
            readonly isDefault?: true | undefined;
            readonly props: {
                [x: string]: import("./metadata.js").PrimitiveProperty | import("./metadata.js").DateProperty | import("./metadata.js").EnumProperty;
            };
            readonly name: string;
            readonly displayName: string;
        };
    } | null | undefined;
    set $dataSource(val: {
        readonly $metadata: {
            readonly type: "dataSource";
            readonly isDefault?: true | undefined;
            readonly props: {
                [x: string]: import("./metadata.js").PrimitiveProperty | import("./metadata.js").DateProperty | import("./metadata.js").EnumProperty;
            };
            readonly name: string;
            readonly displayName: string;
        };
    } | null | undefined);
    /** Wrapper for `$params.includes` */
    get $includes(): string | null | undefined;
    set $includes(val: string | null | undefined);
    get $items(): TItem[];
    set $items(val: TItem[]);
    /** True if the page set in $params.page is greater than 1 */
    get $hasPreviousPage(): boolean;
    /** True if the count retrieved from the last load indicates that there may be pages after the page set in $params.page */
    get $hasNextPage(): boolean;
    /** Decrement the page parameter by 1 if there is a previous page. */
    $previousPage(): void;
    /** Increment the page parameter by 1 if there is a next page. */
    $nextPage(): void;
    get $page(): number;
    set $page(val: number);
    get $pageSize(): number;
    set $pageSize(val: number);
    get $pageCount(): number | null;
    /**
     * A function for invoking the `/load` endpoint, and a set of properties about the state of the last call.
     */
    get $load(): apiClient.ListApiState<[], TModel>;
    /**
     * A function for invoking the `/count` endpoint, and a set of properties about the state of the last call.
     */
    get $count(): apiClient.ItemApiState<[], number>;
    /**
     * Starts auto-loading of the list as changes to its parameters occur.
     * Only usable from Vue setup() or <script setup>. Otherwise, use $startAutoLoad().
     * @param options Options that control the auto-load behavior.
     */
    $useAutoLoad(options?: AutoLoadOptions<this>): void;
    /**
     * Starts auto-loading of the list as changes to its parameters occur.
     * @param vue A Vue instance through which the lifecycle of the watcher will be managed.
     * @param options Options that control the auto-load behavior.
     */
    $startAutoLoad(vue: VueInstance, options?: AutoLoadOptions<this>): void;
    /** Stops auto-loading if it is currently enabled. */
    $stopAutoLoad(): void;
    constructor(
    /** The metadata representing the type of data that this ViewModel handles. */
    $metadata: TModel["$metadata"], 
    /** Instance of an API client for the model through which direct, stateless API requests may be made. */
    $apiClient: TApi);
}
export declare class ServiceViewModel<TMeta extends Service = Service, TApi extends ServiceApiClient<TMeta> = ServiceApiClient<TMeta>> {
    /** The metadata representing the type of data that this ViewModel handles. */
    readonly $metadata: TMeta;
    /** Instance of an API client for the model through which direct, stateless API requests may be made. */
    readonly $apiClient: TApi;
    /** Static lookup of all generated ServiceViewModel types. */
    static typeLookup: ServiceViewModelTypeLookup | null;
    constructor(
    /** The metadata representing the type of data that this ViewModel handles. */
    $metadata: TMeta, 
    /** Instance of an API client for the model through which direct, stateless API requests may be made. */
    $apiClient: TApi);
}
/** Factory for creating new ViewModels from some initial data.
 *
 * For all ViewModels created recursively as a result of creating the root ViewModel,
 * the same ViewModel instance will be returned whenever the exact same `initialData` object is encountered.
 */
export declare class ViewModelFactory {
    private isCleanData;
    private static current;
    private map;
    /** Ask the factory for a ViewModel for the given type and initial data.
     * The instance may be a brand new one, or may be already existing
     * if the same initialData has already been seen.
     */
    get(typeName: string, initialData: any): ViewModel<any, any, any>;
    /** Provide a pre-existing instance to the factory. */
    set(initialData: any, vm: ViewModel): void;
    static scope<TRet>(action: (factory: ViewModelFactory) => TRet, isCleanData: boolean): TRet;
    static get(typeName: string, initialData: any, isCleanData: boolean): ViewModel<any, any, any>;
    private constructor();
}
export declare class ViewModelCollection<T extends ViewModel> extends Array<T> {
    readonly $metadata: ModelCollectionValue | ModelCollectionNavigationProperty | ModelType;
    readonly $parent: ViewModel | ListViewModel;
    $hasLoaded: boolean;
    push(...items: T[]): number;
    splice(start: number, deleteCount?: number, ...items: T[]): T[];
    constructor($metadata: ModelCollectionValue | ModelType, $parent: ViewModel | ListViewModel);
}
type DebounceOptions = {
    /** Time, in milliseconds, to delay. Passed as the second parameter to lodash's `debounce` function. */
    wait?: number;
    /** Additional options to pass to the third parameter of lodash's `debounce` function. */
    debounce?: DebounceSettings;
};
type AutoLoadOptions<TThis> = DebounceOptions & {
    /** A function that will be called before autoloading that can return false to prevent a load. */
    predicate?: (viewModel: TThis) => boolean;
};
type AutoSaveOptions<TThis> = DebounceOptions & ({
    /** A function that will be called before autosaving that can return false to prevent a save. */
    predicate?: (viewModel: TThis) => boolean;
    /** If true, auto-saving will also be enabled for all view models that are
     * reachable from the navigation properties & collections of the current view model. */
    deep?: false;
} | {
    /** A function that will be called before autosaving that can return false to prevent a save. */
    predicate?: (viewModel: ViewModel) => boolean;
    /** If true, auto-saving will also be enabled for all view models that are
     * reachable from the navigation properties & collections of the current view model. */
    deep: true;
});
export interface BulkSaveOptions {
    /** A predicate that will be applied to each modified model
     * to determine if it should be included in the bulk save operation.
     *
     * The predicate is applied before validation (`$hasError`), allowing
     * it to be used to skip over entities that have client validation errors
     * that would otherwise cause the entire bulk save operation to fail.
     * */
    predicate?: (viewModel: ViewModel, action: "save" | "delete") => boolean;
}
/**
 * Dynamically adds getter/setter properties to a class. These properties wrap the properties in its instances' $data objects.
 * @param ctor The class to add wrapper properties to
 * @param metadata The metadata describing the properties to add.
 */
export declare function defineProps<T extends new () => ViewModel<any, any>>(ctor: T, metadata: ModelType): void;
export interface ViewModelTypeLookup {
    [name: string]: new (initialData?: any) => ViewModel;
}
export interface ListViewModelTypeLookup {
    [name: string]: new () => ListViewModel<any, any, any>;
}
export interface ServiceViewModelTypeLookup {
    [name: string]: new () => ServiceViewModel;
}
export type ModelOf<T> = T extends ViewModel<infer TModel> ? TModel : never;
/**
 * Updates the target model with values from the source model.
 * @param target The viewmodel to be updated.
 * @param source The model whose values will be used to perform the update.
 * @param skipDirty If true, only non-dirty props, and related objects, will be updated.
 * Basic properties on target that are dirty will be skipped.
 * @param purgeUnsaved If true, unsaved items lingering in collections that are not in `source` will be removed.
 * @param skipSaving If true, only props not actively being saved, and related objects, will be updated.
 * Basic properties on target that are currently being saved will be skipped.
 */
export declare function updateViewModelFromModel<TViewModel extends ViewModel<Model<ModelType>>>(target: TViewModel, source: Indexable<{}>, skipDirty?: boolean, isCleanData?: boolean, purgeUnsaved?: boolean, skipSaving?: boolean): void;
declare module "@vue/reactivity" {
    interface RefUnwrapBailTypes {
        coalesceViewModels: ViewModel | ListViewModel | ServiceViewModel;
    }
}
declare module "vue" {
    interface GlobalComponents {
    }
    interface RefUnwrapBailTypes {
        coalesceViewModels: ViewModel | ListViewModel | ServiceViewModel;
    }
}
